@@ -4,21 +4,25 @@ use strict;
 use vars qw($VERSION);
 use Carp;
 
-$VERSION = '1.00';
+$VERSION = '2.00';
 
 
 # LOAD FILTERING MODULE...
 use Filter::Util::Call;
 
-my $offset = 0;
 sub __();
 
 # CATCH ATTEMPTS TO CALL case OUTSIDE THE SCOPE OF ANY switch
 
-$::_SWITCH = sub { croak "no switch statement active for case" };
+$::_S_W_I_T_C_H = sub { croak "case statement not in switch block" };
+
+my $offset;
+my $fallthrough;
 
 sub import
 {
+	$fallthrough = grep /\bfallthrough\b/, @_;
+	$offset = (caller)[2]+1;
 	filter_add({}) unless @_>1 && $_[1] ne '__';
 	my $pkg = caller;
 	no strict 'refs';
@@ -27,7 +31,6 @@ sub import
 		*{"${pkg}::$_"} = \&$_;
 	}
 	*{"${pkg}::__"} = \&__ if grep /__/, @_;
-	$offset = (caller)[2]+1;
 	1;
 }
 
@@ -39,17 +42,25 @@ sub unimport
 sub filter
 {
 	my($self) = @_ ;
+	local $Switch::file = (caller)[1];
 
 	my $status = 1;
 	$status = filter_read(10_000);
 	return $status if $status<0;
-	$_ = filter_blocks($_);
+    	$_ = filter_blocks($_,$offset);
 	$_ = "# line $offset\n" . $_ if $offset; undef $offset;
+	# print STDERR $_;
 	return $status;
 }
 
 use Text::Balanced ':ALL';
 
+sub line
+{
+	my ($pretext,$offset) = @_;
+	($pretext=~tr/\n/\n/)+$offset,
+}
+
 sub is_block
 {
 	local $SIG{__WARN__}=sub{die$@};
@@ -61,102 +72,101 @@ sub is_block
 
 sub filter_blocks
 {
-	my $source = shift;
+	my ($source, $line) = @_;
+	return $source unless $source =~ /case|switch/;
 	pos $source = 0;
 	my $text = "";
+	my $casecounter = 1;
 	component: while (pos $source < length $source)
 	{
-		#print STDERR "<<<", substr($source,pos $source,20), ">>>\n";
-		if ($source =~ m/(\G\s*use\s+Switch\b)/gc)
+		if ($source =~ m/(\G\s*use\s+switch\b)/gc)
 		{
-			#print STDERR " <<use switch>>\n";
 			$text .= q{use Switch 'noimport'};
 			next component;
 		}
 		my @pos = Text::Balanced::_match_quotelike(\$source,qr/\s*/,1,1);
 		if (defined $pos[0])
 		{
-			#print STDERR " <<quotelike>>\n";
 			$text .= substr($source,$pos[2],$pos[18]-$pos[2]);
 			next component;
 		}
 		@pos = Text::Balanced::_match_variable(\$source,qr/\s*/);
 		if (defined $pos[0])
 		{
-			#print STDERR " <<variable:",pos($source),">>\n";
 			$text .= substr($source,$pos[0],$pos[4]-$pos[0]);
 			next component;
 		}
 
-		if ($source =~ m/\G(\n*)(\s*)switch/gc)
+		if ($source =~ m/\G(\n*)(\s*)switch\b(?=\s*[(])/gc)
 		{
-			#print STDERR " <<switch>>\n";
-			$text .= $1.'local $::_SWITCH;'.$2.'Switch::switch';
-			$source =~ m/\G(\s*[(]?\s*)(?=%|qw)/gc 
-			   and $text .= $1. '\\';
-			$source =~ m/\G(\s*[(]?\s*)(?=\/)/gc 
-			   and $text .= $1 . 'qr';
-			$source =~ m/\G(\s*[(]?\s*)m\b/gc 
-			   and $text .= $1 . 'qr';
-			@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/\{/,qr/\}/,qr/\{/,qr/\}/,undef);
-			if (@pos)
-			{
-				my $code = substr($source,$pos[0],$pos[4]-$pos[0]);
-				$text .= " sub" if is_block $code;
-				$text .= " " . filter_blocks($code);
-				next component;
-			}
+			$text .= $1.$2.'S_W_I_T_C_H: while (1) ';
+			@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/\(/,qr/\)/,qr/\{/,qr/\}/,undef) 
+			or do {
+				die "Bad switch statement (problem in the parentheses?) near $Switch::file line ", line(substr($source,0,pos $source),$line), "\n";
+			};
+			my $arg = filter_blocks(substr($source,$pos[0],$pos[4]-$pos[0]),line(substr($source,0,$pos[0]),$line));
+			$arg =~ s {^\s*[(]\s*%}   { ( \\\%}	||
+			$arg =~ s {^\s*[(]\s*m\b} { ( qr}	||
+			$arg =~ s {^\s*[(]\s*/}   { ( qr/}	||
+			$arg =~ s {^\s*[(]\s*qw}  { ( \\qw};
+			@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/\{/,qr/\}/,qr/\{/,qr/\}/,undef)
+			or do {
+				die "Bad switch statement (problem in the code block?) near $Switch::file line ", line(substr($source,0, pos $source), $line), "\n";
+			};
+			my $code = filter_blocks(substr($source,$pos[0],$pos[4]-$pos[0]),line(substr($source,0,$pos[0]),$line));
+			$code =~ s/{/{ local \$::_S_W_I_T_C_H; Switch::switch $arg;/;
+			$text .= $code . 'continue {last}';
+			next component;
 		}
-		elsif ($source =~ m/\G(\s*)(case)(?!\s*=>)/gc)
+		elsif ($source =~ m/\G(\s*)(case\b)(?!\s*=>)/gc)
 		{
-			#print STDERR " <<case>>\n";
 			$text .= $1."if (Switch::case";
-			@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/\{/,qr/\}/,qr/\{/,qr/\}/,undef);
-			if (@pos)
-			{
+			if (@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/\{/,qr/\}/,qr/\{/,qr/\}/,undef)) {
 				my $code = substr($source,$pos[0],$pos[4]-$pos[0]);
 				$text .= " sub" if is_block $code;
-				$text .= " " . filter_blocks($code) . ")";
-				next component;
+				$text .= " " . filter_blocks($code,line(substr($source,0,$pos[0]),$line)) . ")";
 			}
-			@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/[[(]/,qr/[])]/,qr/[[({]/,qr/[])}]/,undef);
-			if (@pos)
-			{
-				my $code = filter_blocks(substr($source,$pos[0],$pos[4]-$pos[0]));
+			elsif (@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/[[(]/,qr/[])]/,qr/[[({]/,qr/[])}]/,undef)) {
+				my $code = filter_blocks(substr($source,$pos[0],$pos[4]-$pos[0]),line(substr($source,0,$pos[0]),$line));
 				$code =~ s {^\s*[(]\s*%}   { ( \\\%}	||
 				$code =~ s {^\s*[(]\s*m\b} { ( qr}	||
 				$code =~ s {^\s*[(]\s*/}   { ( qr/}	||
 				$code =~ s {^\s*[(]\s*qw}  { ( \\qw};
 				$text .= " $code)";
-				next component;
 			}
-			@pos = Text::Balanced::_match_quotelike(\$source,qr/\s*/,1,1);
-			if (@pos)
-			{
+			elsif ( @pos = Text::Balanced::_match_quotelike(\$source,qr/\s*/,1,1)) {
 				my $code = substr($source,$pos[2],$pos[18]-$pos[2]);
-				#print STDERR "[[[$code]]]\n";
-				$code = filter_blocks($code);
+				$code = filter_blocks($code,line(substr($source,0,$pos[2]),$line));
 				$code =~ s {^\s*m}  { qr}	||
 				$code =~ s {^\s*/}  { qr/}	||
 				$code =~ s {^\s*qw} { \\qw};
 				$text .= " $code)";
-				next component;
 			}
-			if ($source =~ m/\G\s*(([^\$\@{])[^\$\@{]*)(?=\s*{)/gc)
-			{
-				my $code = filter_blocks($1);
+			elsif ($source =~ m/\G\s*(([^\$\@{])[^\$\@{]*)(?=\s*{)/gc) {
+				my $code = filter_blocks($1,line(substr($source,0,pos $source),$line));
 				$text .= ' \\' if $2 eq '%';
 				$text .= " $code)";
-				next component;
+			}
+			else {
+				die "Bad case statement (invalid case value?) near $Switch::file line ", line(substr($source,0,pos $source), $line), "\n";
 			}
 
-			my @caller = caller 1;
-			die "Bad case statement: $@ near ", $caller[1], " line ",
-			      (substr($source,0,pos $source)=~tr/\n/\n/)+1+$caller[2],
-			      "\n";
+			@pos = Text::Balanced::_match_codeblock(\$source,qr/\s*/,qr/\{/,qr/\}/,qr/\{/,qr/\}/,undef)
+			or do {
+				if ($source =~ m/\G\s*(?=([};]|\Z))/gc) {
+					$casecounter++;
+					next component;
+				}
+				die "Bad case statement (problem in the code block?) near $Switch::file line ", line(substr($source,0,pos $source),$line), "\n";
+			};
+			my $code = filter_blocks(substr($source,$pos[0],$pos[4]-$pos[0]),line(substr($source,0,$pos[0]),$line));
+			$code =~ s/}(?=\s*\Z)/;last S_W_I_T_C_H }/
+				unless $fallthrough;
+			$text .= "{ while (1) $code continue { goto C_A_S_E_$casecounter } last S_W_I_T_C_H; C_A_S_E_$casecounter: }";
+			$casecounter++;
+			next component;
 		}
 
-		#print STDERR " <<other>>\n";
 		$source =~ m/\G(\s*(\w+|#.*\n|\W))/gc;
 		$text .= $1;
 	}
@@ -204,7 +214,7 @@ sub switch(;$)
 	
 	if ($s_ref eq 'CODE')
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    return $s_val == $c_val  if ref $c_val eq 'CODE';
 			    return $s_val->(@$c_val) if ref $c_val eq 'ARRAY';
@@ -213,7 +223,7 @@ sub switch(;$)
 	}
 	elsif ($s_ref eq "" && (~$s_val&$s_val) eq 0)	# NUMERIC SCALAR
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    my $c_ref = ref $c_val;
 			    return $s_val == $c_val 	if $c_ref eq ""
@@ -231,7 +241,7 @@ sub switch(;$)
 	}
 	elsif ($s_ref eq "")				# STRING SCALAR
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    my $c_ref = ref $c_val;
 			    return $s_val eq $c_val 	if $c_ref eq "";
@@ -247,7 +257,7 @@ sub switch(;$)
 	}
 	elsif ($s_ref eq 'ARRAY')
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    my $c_ref = ref $c_val;
 			    return in($s_val,[$c_val]) 	if $c_ref eq "";
@@ -264,7 +274,7 @@ sub switch(;$)
 	}
 	elsif ($s_ref eq 'Regexp')
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    my $c_ref = ref $c_val;
 			    return $c_val=~/s_val/ 	if $c_ref eq "";
@@ -280,7 +290,7 @@ sub switch(;$)
 	}
 	elsif ($s_ref eq 'HASH')
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    my $c_ref = ref $c_val;
 			    return $s_val->{$c_val} 	if $c_ref eq "";
@@ -296,7 +306,7 @@ sub switch(;$)
 	}
 	elsif ($s_ref eq 'Switch')
 	{
-		$::_SWITCH =
+		$::_S_W_I_T_C_H =
 		      sub { my $c_val = $_[0];
 			    return $s_val == $c_val  if ref $c_val eq 'Switch';
 			    return $s_val->call(@$c_val)
@@ -311,7 +321,7 @@ sub switch(;$)
 	return 1;
 }
 
-sub case($) { $::_SWITCH->(@_); }
+sub case($) { $::_S_W_I_T_C_H->(@_); }
 
 # IMPLEMENT __
 
@@ -443,14 +453,14 @@ Switch - A switch statement for Perl
 
 =head1 VERSION
 
-This document describes version 1.00 of Switch,
-released August  7, 2000.
+This document describes version 2.00 of Switch,
+released January  8, 2001.
 
 =head1 SYNOPSIS
 
 	use Switch;
 
-	switch ($val);
+	switch ($val) {
 
 		case 1		{ print "number 1" }
 		case "a"	{ print "string a" }
@@ -462,16 +472,20 @@ released August  7, 2000.
 		case (\%hash)	{ print "entry in hash" }
 		case (\&sub)	{ print "arg to subroutine" }
 		else		{ print "previous case not true" }
+	}
 
 =head1 BACKGROUND
 
+[Skip ahead to L<"DESCRIPTION"> if you don't care about the whys
+and wherefores of this control structure]
+
 In seeking to devise a "Swiss Army" case mechanism suitable for Perl,
 it is useful to generalize this notion of distributed conditional
 testing as far as possible. Specifically, the concept of "matching"
 between the switch value and the various case values need not be
 restricted to numeric (or string or referential) equality, as it is in other 
 languages. Indeed, as Table 1 illustrates, Perl
-offers at least sixteen different ways in which two values could
+offers at least eighteen different ways in which two values could
 generate a match.
 
 	Table 1: Matching a switch value ($s) with a case value ($c)
@@ -483,6 +497,10 @@ generate a match.
         number  same    numeric or referential  match if $s == $c;
         or ref          equality
 
+	object  method	result of method call   match if $s->$c();
+	ref     name 				match if defined $s->$c();
+		or ref
+
         other   other   string equality         match if $s eq $c;
         non-ref non-ref
         scalar  scalar
@@ -516,7 +534,7 @@ generate a match.
         ref     ref     return value truth      match if $s->(@$c);
 
 
-In reality, Table 1 covers 29 alternatives, because only the equality and
+In reality, Table 1 covers 31 alternatives, because only the equality and
 intersection tests are commutative; in all other cases, the roles of
 the C<$s> and C<$c> variables could be reversed to produce a
 different test. For example, instead of testing a single hash for
@@ -535,55 +553,63 @@ the numerous possible combinations of switch and case values described above.
 
 The module augments the standard Perl syntax with two new control
 statements: C<switch> and C<case>. The C<switch> statement takes a
-single optional scalar argument of any type (including a subroutine
-reference specified as a simple block). If the argument is a variable,
-it must be enclosed in parentheses; otherwise the parentheses are
-optional. If the argument is omitted, C<$_> is used instead. Regardless
-of the origin or type of argument, C<switch> stores its value as the
-current switch value in the (localized) control variable C<${::SWITCH}>.
+single scalar argument of any type, specified in parentheses.
+C<switch> stores this value as the
+current switch value in a (localized) control variable.
+The value is followed by a block which may contain one or more
+Perl statements (including the C<case> statement described below).
+The block is unconditionally executed once the switch value has
+been cached.
 
-The C<case> statement takes a single scalar argument (again, in mandatory
-parentheses if it's a variable) and selects the
-appropriate type of matching between that argument and the current value
-of C<${^::SWITCH}>. The type of matching used is determined by the
-respective types of the C<${::SWITCH}> value and the C<case> argument, as
-specified in Table 1. If the match is successful, the block associated
-with the C<case> statement is executed.
+A C<case> statement takes a single scalar argument (in mandatory
+parentheses if it's a variable; otherwise the parens are optional) and
+selects the appropriate type of matching between that argument and the
+current switch value. The type of matching used is determined by the
+respective types of the switch value and the C<case> argument, as
+specified in Table 1. If the match is successful, the mandatory
+block associated with the C<case> statement is executed.
 
-In all other respects, the C<case> statement is semantically identical
+In most other respects, the C<case> statement is semantically identical
 to an C<if> statement. For example, it can be followed by an C<else>
 clause, and can be used as a postfix statement qualifier. 
 
+However, when a C<case> block has been executed control is automatically
+transferred to the statement after the immediately enclosing C<switch>
+block, rather than to the next statement within the block. In other
+words, the success of any C<case> statement prevents other cases in the
+same scope from executing. But see L<"Allowing fall-through"> below.
+
 Together these two new statements provide a fully generalized case
 mechanism:
 
-        use switch;
+        use Switch;
 
         # AND LATER...
 
         %special = ( woohoo => 1,  d'oh => 1 );
 
         while (<>) {
-                switch ($_);
+            switch ($_) {
 
-                case %special  { print "homer\n"; next }       # if $special{$_}
-                case /a-z/i    { print "alpha\n"; next }       # if $_ =~ /a-z/i
-                case [1..9]    { print "small num\n"; next }   # if $_ in [1..9]
+                case %special  { print "homer\n"; }       # if $special{$_}
+                case /a-z/i    { print "alpha\n"; }       # if $_ =~ /a-z/i
+                case [1..9]    { print "small num\n"; }   # if $_ in [1..9]
 
-                case { $_[0] >= 10 } {    		       # if $_ >= 10
-                        my $age = <>;
-                        switch { $_[0] < $age };
+                case { $_[0] >= 10 } {                    # if $_ >= 10
+                    my $age = <>;
+                    switch (sub{ $_[0] < $age } ) {
 
-                        case 20  { print "teens\n"; next }     # if 20 < $age
-                        case 30  { print "twenties\n"; next }  # if 30 < $age
-                        else     { print "history\n"; next }
+                        case 20  { print "teens\n"; }     # if 20 < $age
+                        case 30  { print "twenties\n"; }  # if 30 < $age
+                        else     { print "history\n"; }
+                    }
                 }
 
-                print "must be punctuation\n" case /\W/;       # if $_ ~= /\W/
+                print "must be punctuation\n" case /\W/;  # if $_ ~= /\W/
         }
 
 Note that C<switch>es can be nested within C<case> (or any other) blocks,
-and a series of case statements can try different types of matches
+and a series of C<case> statements can try different types of matches
 -- hash membership, pattern match, array intersection, simple equality,
 etc. -- against the same switch value.
 
@@ -592,13 +618,87 @@ useful for aggregating integral cases:
 
         sub classify_digit
         {
-                switch ($_[0]);  case 0            { return 'zero' }
+                switch ($_[0]) { case 0            { return 'zero' }
                                  case [2,4,6,8]    { return 'even' }
                                  case [1,3,4,7,9]  { return 'odd' }
                                  case /[A-F]/i     { return 'hex' }
+                               }
         }
 
 
+=head2 Allowing fall-through
+
+Fall-though (trying another case after one has already succeeded)
+is usually a Bad Idea in a switch statement. However, this
+is Perl, not a police state, so there I<is> a way to do it, if you must.
+
+If a C<case> block executes an untargetted C<next>, control is
+immediately transferred to the statement I<after> the C<case> statement
+(i.e. usually another case), rather than out of the surrounding
+C<switch> block.
+
+For example:
+
+        switch ($val) {
+                case 1      { handle_num_1(); next }    # and try next case...
+                case "1"    { handle_str_1(); next }    # and try next case...
+                case [0..9] { handle_num_any(); }       # and we're done
+                case /\d/   { handle_dig_any(); next }  # and try next case...
+                case /.*/   { handle_str_any(); next }  # and try next case...
+        }
+
+If $val held the number C<1>, the above C<switch> block would call the
+first three C<handle_...> subroutines, jumping to the next case test
+each time it encountered a C<next>. After the thrid C<case> block
+was executed, control would jump to the end of the enclosing
+C<switch> block.
+
+On the other hand, if $val held C<10>, then only the last two C<handle_...>
+subroutines would be called.
+
+Note that this mechanism allows the notion of I<conditional fall-through>.
+For example:
+
+        switch ($val) {
+                case [0..9] { handle_num_any(); next if $val < 7; }
+                case /\d/   { handle_dig_any(); }
+        }
+
+If an untargetted C<last> statement is executed in a case block, this
+immediately transfers control out of the enclosing C<switch> block
+(in other words, there is an implicit C<last> at the end of each
+normal C<case> block). Thus the previous example could also have been
+written:
+
+        switch ($val) {
+                case [0..9] { handle_num_any(); last if $val >= 7; next; }
+                case /\d/   { handle_dig_any(); }
+        }
+
+
+=head2 Automating fall-through
+
+In situations where case fall-through should be the norm, rather than an
+exception, an endless succession of terminal C<next>s is tedious and ugly.
+Hence, it is possible to reverse the default behaviour by specifying
+the string "fallthrough" when importing the module. For example, the 
+following code is equivalent to the first example in L<"Allowing fall-through">:
+
+        use Switch 'fallthrough';
+
+        switch ($val) {
+                case 1      { handle_num_1(); }
+                case "1"    { handle_str_1(); }
+                case [0..9] { handle_num_any(); last }
+                case /\d/   { handle_dig_any(); }
+                case /.*/   { handle_str_any(); }
+        }
+
+Note the explicit use of a C<last> to preserve the non-fall-through
+behaviour of the third case.
+
+
+
 =head2 Higher-order Operations
 
 One situation in which C<switch> and C<case> do not provide a good
@@ -606,24 +706,25 @@ substitute for a cascaded C<if>, is where a switch value needs to
 be tested against a series of conditions. For example:
 
         sub beverage {
-                switch shift;
+            switch (shift) {
 
-                case { $_[0] < 10 }  { return 'milk' }
-                case { $_[0] < 20 }  { return 'coke' }
-                case { $_[0] < 30 }  { return 'beer' }
-                case { $_[0] < 40 }  { return 'wine' }
-                case { $_[0] < 50 }  { return 'malt' }
-                case { $_[0] < 60 }  { return 'Moet' }
-                else                 { return 'milk' }
+                case sub { $_[0] < 10 }  { return 'milk' }
+                case sub { $_[0] < 20 }  { return 'coke' }
+                case sub { $_[0] < 30 }  { return 'beer' }
+                case sub { $_[0] < 40 }  { return 'wine' }
+                case sub { $_[0] < 50 }  { return 'malt' }
+                case sub { $_[0] < 60 }  { return 'Moet' }
+                else                     { return 'milk' }
+            }
         }
 
-The need to specify each condition as an implicit subroutine block is
-tiresome. To overcome this, when importing Switch.pm, users may elect to
-also import a special "placeholder" subroutine named C<__> [sic]. This
-subroutine converts (almost) any expression in which it appears to a
-reference to a higher-order function. That is, the expression:
+The need to specify each condition as a subroutine block is tiresome. To
+overcome this, when importing Switch.pm, a special "placeholder"
+subroutine named C<__> [sic] may also be imported. This subroutine
+converts (almost) any expression in which it appears to a reference to a
+higher-order function. That is, the expression:
 
-        use switch '__';
+        use Switch '__';
 
         __ < 2 + __
 
@@ -662,6 +763,11 @@ This boolean expression is inevitably true, since both references are
 non-false. Fortunately, the overloaded C<'bool'> operator catches this
 situation and flags it as a error. 
 
+=head1 DEPENDENCIES
+
+The module is implemented using Filter::Util::Call and Text::Balanced
+and requires both these modules to be installed. 
+
 =head1 AUTHOR
 
 Damian Conway (damian@conway.org)
